/*********************************************************************
 *                
 * Copyright (C) 1999, 2000, 2001,  Karlsruhe University
 *                
 * File path:     asm-x86.S
 * Description:   X86-specifc glue for the kernel debugger.
 *                
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *                
 * asm-x86.S,v 1.11.2.2 2001/12/11 12:27:44 skoglund Exp
 *                
 ********************************************************************/
#include <config.h>
#include <x86/config.h>
#include <x86/cpu.h>
#include <tcb_layout.h>

#if (__GNUC__ >= 3)
# define KDEBUG_INIT   _Z11kdebug_initv
# define KMEM_ALLOC    _Z10kmem_allocj
# define TRACE_PF_HANDLER	_Z16tracs_pf_handlerjjjjj
# define TRACE_IPC_HANDLER	_Z17trace_ipc_handllerjjjjjjj
# define PMC_OVERFLOW_HANDLER	_Z20pmc_overflow_handlerv
# define TRACE_EXCEPTION_HANDLER	_Z17exception_handlerP17exception_frame_t
#else
# define KDEBUG_INIT   kdebug_init__Fv
# define KMEM_ALLOC    kmem_alloc__FUi
# define TRACE_PF_HANDLER	trace_pf_handler__FUiUiUiUiUi
# define TRACE_IPC_HANDLER	trace_ipc_handler__FUiUiUiUiUiUiUi
# define PMC_OVERFLOW_HANDLER	pmc_overflow_handler__Fv
# define TRACE_EXCEPTION_HANDLER exception_handler__FP17exception_frame_t
#endif

/*
 * Define to keep the old copy of the stack contents.  Useful if one 
 * catches faults inside KDB.
 */
//#define KEEP_OLD_STACK_CONTENTS

#if defined(CONFIG_X86_P4_BTS)
#  define DEBUGCTL_MASK	(~0x0d)
#else
#  define DEBUGCTL_MASK	(~0x01)
#endif

#if defined(CONFIG_KEEP_LAST_BRANCHES) || defined(CONFIG_X86_P4_BTS)
#define clr_dbgctl()				  \
	mov	$IA32_DEBUGCTL, %ecx		; \
	rdmsr					; \
	andl	$DEBUGCTL_MASK, %eax		; \
	wrmsr
#define set_dbgctl()				  \
	mov	$IA32_DEBUGCTL, %ecx		; \
	rdmsr					; \
	movl	x86_debugctl_state, %ebx	; \
	orl	%ebx, %eax			; \
	wrmsr
#else
#define clr_dbgctl()
#define set_dbgctl()
#endif

#define ENTRY(name) \
	.globl name; \
	.align 4; \
	name##:


	
#define ke(text) \
	int	$3		; \
	jmp	9f		; \
	.ascii	text		; \
9:

	.section .kdebug

ENTRY(int_1)
	pushl	$0x0
	pusha
	pushl	%ds
	pushl	%es
	pushl	$1
	jmp	0f

ENTRY(int_2)
	pushl	$0x0
	pusha
	pushl	%ds
	pushl	%es
	pushl	$2
	jmp	0f


ENTRY(int_3)
	pushl	$0x0
	pusha
	pushl	%ds
	pushl	%es
	pushl	$3

0:
#if defined(CONFIG_ENABLE_SMALL_AS)
	mov	$X86_KDS, %eax
	mov	%eax, %ds
	mov	%eax, %es
#endif

	clr_dbgctl()

	cmpl	$KDEBUG_INIT, %ss:(kernel_info_page + 0x10)
	je	2f
	call	*%ss:(kernel_info_page + 0x14)
2:	
	mov	%esp, %ss:_kdebug_saved_esp
	mov	$_kdebug_stack_top, %esp
	pushl	_kdebug_saved_esp
#if defined(KEEP_OLD_STACK_CONTENTS)
	cmpl	$_kdebug_stack_bottom, (%esp)
	jl	4f
	movl	$_kdebug_stack_bottom, %esi
	movl	$_kdebug_old_stack, %edi
	movl	$256, %ecx
	rep
	movsl	%ds:(%esi), %es:(%edi)
	addl	$(_kdebug_old_stack-_kdebug_stack_bottom), (%esp)
4:
#endif
	call	*%ss:(kernel_info_page + 0x14)
	mov	%ss:_kdebug_saved_esp, %esp

	set_dbgctl()
	addl	$4, %esp
	popl	%es
	popl	%ds
	popa
	addl	$4,%esp
	iret 

	

ENTRY(grabpage__Fv)
	/* save the registers the callee doesn't save itself */
	pushl	%ecx
	pushl	%edx
	pushl	$4096
	call	KMEM_ALLOC
	subl	$KERNEL_OFFSET, %eax
	addl	$4,%esp
	popl	%edx
	popl	%ecx
	ret

_kdebug_counter:
	.long	0

	.section .kdebug-bss
	.global _kdebug_stack_bottom
	.global _kdebug_stack_top
	.global _kdebug_saved_esp
	.align 16
_kdebug_stack_bottom:
	.space	1024
_kdebug_stack_top:

#if defined(KEEP_OLD_STACK_CONTENTS)
_kdebug_old_stack:	
	.space	1024
#endif

#if defined(CONFIG_KEEP_LAST_BRANCHES) || defined(CONFIG_X86_P4_BTS)
	.global x86_debugctl_state
x86_debugctl_state:
	.long	0
#endif

_kdebug_saved_esp:
	.long	0

/* the assembler stubs for the new debugger */


