/*********************************************************************
 *                
 * Copyright (C) 2001,  Karlsruhe University
 *                
 * File path:     input.c
 * Description:   Various kdebug input functions.
 *                
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *                
 * input.c,v 1.14.2.1 2001/12/11 12:27:44 skoglund Exp
 *                
 ********************************************************************/
#include <universe.h>

#include "kdebug.h"



tcb_t *kdebug_get_thread(tcb_t *current)
{
    l4_threadid_t tid;

    printf("Thread no [current]: ");

    tid.raw = kdebug_get_hex((dword_t) current, "current");
    if ( tid.raw < TCB_AREA ) tid.raw = (dword_t) tid_to_tcb(tid);
    else tid.raw = (tid.raw & L4_TCB_MASK);

    return (tcb_t *) tid.raw;
}


tcb_t *kdebug_get_task(tcb_t *current)
{
    l4_threadid_t tid;

    printf("Task no [current]: ");

    tid.raw = kdebug_get_hex((dword_t) current, "current");
    if ( tid.raw < (1 << L4_X0_TASKID_BITS) )
	tid = (l4_threadid_t) { x0id: {0, 0, tid.raw, 0} };
    if ( tid.raw < TCB_AREA ) tid.raw = (dword_t) tid_to_tcb(tid);
    tid.x0id.thread = 0;
    tid.raw = (tid.raw & L4_TCB_MASK);

    return (tcb_t *) tid.raw;
}

ptr_t kdebug_get_pgtable(tcb_t *current, char *str)
{
    tcb_t *tcb;
    ptr_t pgtab;
    int first_time;

    if ( str )
	printf("%s", str);
    
    pgtab = (ptr_t) kdebug_get_hex((dword_t) get_current_pgtable(), "current");

    if ( (dword_t) pgtab < (1 << L4_X0_TASKID_BITS) )
    {
	/* Seems like a task id.  Search for valid thread. */
	for ( tcb = current, first_time = 1;
	      tcb != current || first_time;
	      tcb = tcb->present_next, first_time = 0 )
	    if ( tcb->myself.x0id.task == (dword_t) pgtab)
	    {
		pgtab = tcb->space->pagedir_phys ();
		break;
	    }
    }
    else if ( (dword_t) pgtab < TCB_AREA )
    {
	/* Might be a thread id.  Search for match. */
	for ( tcb = current, first_time = 1;
	      tcb != current || first_time;
	      tcb = tcb->present_next, first_time = 0 )
	    if ( tcb->myself.raw == (dword_t) pgtab)
	    {
		pgtab = tcb->space->pagedir_phys ();
		break;
	    }
    }
    else if ( ((dword_t) pgtab >= TCB_AREA &&
	       (dword_t) pgtab < (TCB_AREA + TCB_AREA_SIZE)) )
	/* Pointer into TCB area. */
	pgtab = ((tcb_t *) ((dword_t) pgtab & L4_TCB_MASK))->space->pagedir_phys ();
    else
	/* Hopefully a valid page table address. */
	;

    return pgtab;
}

dword_t kdebug_get_hex(dword_t def, const char *str)
{
    dword_t val = 0;
    char t, c;
    int i = 0;


    while ( (t = c = getc()) != '\r' )
    {
	switch(c)
	{
	case '0' ... '9': c -= '0'; break;
	case 'a' ... 'f': c -= 'a' - 'A';
	case 'A' ... 'F': c = c - 'A' + 10; break;
	case 8:  val >>= 4; i--; putc(8); putc(' '); putc(8); continue;
	case 'x': if ( val == 0 && i == 1 ) { i--; putc('x'); }
	default: continue;
	}
	val <<= 4;
	val += c;
	putc(t);
	if ( ++i == 8 )
	    break;
    }
    if ( i == 0 )
    {
	/* No value given.  Use default. */
	val = def;
	if ( str ) printf("%s", str);
	else printf("%x", val);
    }

    putc('\n');
    return val;
}


dword_t kdebug_get_dec(dword_t def, const char *str)
{
    dword_t val = 0, no_input = 1;
    char c;

    while ( (c = getc()) != '\r' )
    {
	switch(c)
	{
	case '0' ... '9':
	    val = val * 10 + (c - '0');
	    break;
	default:
	    continue;
	}
	putc(c);
	no_input = 0;
    }

    if ( no_input )
    {
	/* No value given.  Use default. */
	val = def;
	if ( str ) printf("%s", str);
	else printf("%d", val);
    }

    putc('\n');
    return val;
}



static char *get_key_string(char key, const char *choices) L4_SECT_KDEBUG;

char kdebug_get_choice(const char *str, const char *choices,
		       char def, int no_newline)
{
    char *p, c;

    printf("%s (%s)", str, choices);
    if ( (p = get_key_string(def, choices)) )
    {
	printf(" [");
	for (; *p != '/' && *p != 0; p++ )
	    putc(*p >= 'A' && *p <= 'Z' ? (*p + ('a'-'A')) : *p);
	putc(']');
    }
    printf(": ");

    for (;;)
    {
	c = getc();
	p = get_key_string(c == '\r' ? def : c, choices);
	for (; p && *p != '/' && *p != 0; p++ )
	    putc(*p >= 'A' && *p <= 'Z' ? (*p + ('a'-'A')) : *p);
	if ( p || c == '\r' )
	    break;
    }

    if ( ! no_newline )
	putc('\n');

    return (c == '\r') ? def : c;
}

static char *get_key_string(char key, const char *choices)
{
    const char *p;

    if ( key == 0 ) return NULL;

    for ( p = choices; *p; p++ )
    {
	if ( (*p >= 'A' && *p <= 'Z') && (*p + ('a'-'A')) == key )
	    break;
	if ( (p == choices || p[-1] == '/') && (p[1] == 0 || p[1] == '/') &&
	     (*p == key) )
	    break;
    }

    if ( *p == 0 ) return NULL;
    while ( p > choices && p[-1] != '/' ) p--;
    return (char *) p;
}



