/*********************************************************************
 *                
 * Copyright (C) 2001,  Karlsruhe University
 *                
 * File path:     arm/resources.c
 * Description:   Thread resource management for ARM.
 *                
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *                
 * resources.c,v 1.1 2001/12/07 19:05:43 skoglund Exp
 *                
 ********************************************************************/

#include <universe.h>
#include <thread.h>
#include <tracepoints.h>

void save_resources (tcb_t *current, tcb_t *dest)
{
    TRACEPOINT	(SAVE_RESOURCES, printf ("save_resources: tcb=%p  rc=%p\n",
					 current, current->resources));

    if (current->resources | TR_IPC_MEM)
    {
	ptr_t pdir = current->space->pagedir ();
	pdir[MEM_COPYAREA1 >> PAGEDIR_BITS] = 0;
	pdir[(MEM_COPYAREA1 >> PAGEDIR_BITS) + 1] = 0;

	pdir[MEM_COPYAREA2 >> PAGEDIR_BITS] = 0;
	pdir[(MEM_COPYAREA2 >> PAGEDIR_BITS) + 1] = 0;

	if ((dest == get_idle_tcb ()) || same_address_space (current, dest))
	    flush_tlb ();
    }
	
    current->resources = 0;
}

void purge_resources (tcb_t * tcb)
{
    TRACEPOINT (PURGE_RESOURCES, printf("purge_resources: tcb=%p\n", tcb));
}


void load_resources (tcb_t *current)
{
    TRACEPOINT (LOAD_RESOURCES, printf ("load_resources: tcb=%p  rc=%p\n",
					current, current->resources));
}


void init_resources (tcb_t * tcb)
{ 
}

void free_resources (tcb_t * tcb)
{
}

